<?php

namespace Oro\Bundle\ReportBundle\Controller;

use Doctrine\DBAL\Types\Type;
use Oro\Bundle\ChartBundle\Model\ChartOptionsBuilder;
use Oro\Bundle\ChartBundle\Model\ChartViewBuilder;
use Oro\Bundle\DashboardBundle\Helper\DateHelper;
use Oro\Bundle\DataGridBundle\Datagrid\DatagridInterface;
use Oro\Bundle\DataGridBundle\Datagrid\Manager;
use Oro\Bundle\DataGridBundle\Extension\Pager\PagerInterface;
use Oro\Bundle\EntityBundle\Provider\EntityProvider;
use Oro\Bundle\EntityConfigBundle\Provider\ConfigProvider;
use Oro\Bundle\FeatureToggleBundle\Checker\FeatureChecker;
use Oro\Bundle\QueryDesignerBundle\QueryDesigner\Manager as QueryDesignerManager;
use Oro\Bundle\ReportBundle\Entity\Report;
use Oro\Bundle\ReportBundle\Entity\ReportType;
use Oro\Bundle\ReportBundle\Form\Handler\ReportHandler;
use Oro\Bundle\ReportBundle\Form\Type\ReportType as ReportFormType;
use Oro\Bundle\ReportBundle\Grid\ReportDatagridConfigurationProvider;
use Oro\Bundle\SecurityBundle\Annotation\Acl;
use Oro\Bundle\SecurityBundle\Annotation\AclAncestor;
use Oro\Bundle\SegmentBundle\Provider\EntityNameProvider;
use Oro\Bundle\UIBundle\Route\Router;
use Sensio\Bundle\FrameworkExtraBundle\Configuration\Template;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * Covers the CRUD functionality and the additional operation clone for the Report entity.
 */
class ReportController extends AbstractController
{
    /**
     * {@inheritdoc}
     */
    public static function getSubscribedServices()
    {
        return array_merge(parent::getSubscribedServices(), [
            'oro_entity_config.provider.entity' => ConfigProvider::class,
            EntityProvider::class,
            EntityNameProvider::class,
            QueryDesignerManager::class,
            Manager::class,
            ReportDatagridConfigurationProvider::class,
            ChartViewBuilder::class,
            ChartOptionsBuilder::class,
            TranslatorInterface::class,
            ReportHandler::class,
            Router::class,
            FeatureChecker::class
        ]);
    }

    /**
     * @Route("/view/{id}", name="oro_report_view", requirements={"id"="\d+"})
     * @Acl(
     *      id="oro_report_view",
     *      type="entity",
     *      permission="VIEW",
     *      class="OroReportBundle:Report"
     * )
     *
     * @param Report $entity
     * @return Response
     */
    public function viewAction(Report $entity)
    {
        $this->checkReport($entity);
        $this->get(EntityNameProvider::class)->setCurrentItem($entity);

        $reportGroup = $this->get('oro_entity_config.provider.entity')
            ->getConfig($entity->getEntity())
            ->get('plural_label');
        $parameters  = [
            'entity'      => $entity,
            'reportGroup' => $reportGroup
        ];

        $reportType = $entity->getType()->getName();
        if ($reportType === ReportType::TYPE_TABLE) {
            $gridName = $entity::GRID_PREFIX . $entity->getId();

            if ($this->get(ReportDatagridConfigurationProvider::class)->isReportValid($gridName)) {
                $parameters['gridName'] = $gridName;

                $datagrid = $this->get(Manager::class)->getDatagrid(
                    $gridName,
                    [PagerInterface::PAGER_ROOT_PARAM => [PagerInterface::DISABLED_PARAM => true]]
                );

                $chartOptions = $this->get(ChartOptionsBuilder::class)->buildOptions(
                    $entity->getChartOptions(),
                    $datagrid->getConfig()->toArray()
                );

                if (!empty($chartOptions)) {
                    $chartOptions = $this->processChartOptions($datagrid, $chartOptions);

                    $parameters['chartView'] = $this->get(ChartViewBuilder::class)
                        ->setDataGrid($datagrid)
                        ->setOptions($chartOptions)
                        ->getView();
                }
            }
        }

        return $this->render(
            sprintf('OroReportBundle:Report:%s/view.html.twig', strtolower($reportType)),
            $parameters
        );
    }

    /**
     * @Route("/view/{gridName}", name="oro_report_view_grid", requirements={"gridName"="[-\w]+"})
     *
     * @Template
     * @Acl(
     *      id="oro_report_view",
     *      type="entity",
     *      permission="VIEW",
     *      class="OroReportBundle:Report"
     * )
     *
     * @param string $gridName
     * @return array
     */
    public function viewFromGridAction($gridName)
    {
        $configuration = $this->get(Manager::class)->getConfigurationForGrid($gridName);
        $pageTitle = isset($configuration['pageTitle']) ? $configuration['pageTitle'] : $gridName;

        return [
            'pageTitle' => $this->get(TranslatorInterface::class)->trans($pageTitle),
            'gridName'  => $gridName,
        ];
    }

    /**
     * @Route("/create", name="oro_report_create")
     * @Template("OroReportBundle:Report:update.html.twig")
     * @Acl(
     *      id="oro_report_create",
     *      type="entity",
     *      permission="CREATE",
     *      class="OroReportBundle:Report"
     * )
     */
    public function createAction()
    {
        return $this->update(new Report());
    }

    /**
     * @Route("/update/{id}", name="oro_report_update", requirements={"id"="\d+"})
     *
     * @Template
     * @Acl(
     *      id="oro_report_update",
     *      type="entity",
     *      permission="EDIT",
     *      class="OroReportBundle:Report"
     * )
     *
     * @param Report $entity
     * @return array
     */
    public function updateAction(Report $entity)
    {
        $this->checkReport($entity);

        return $this->update($entity);
    }

    /**
     * @Route("/clone/{id}", name="oro_report_clone", requirements={"id"="\d+"})
     * @Template("OroReportBundle:Report:update.html.twig")
     * @AclAncestor("oro_report_create")
     *
     * @param Report $entity
     * @return array
     */
    public function cloneAction(Report $entity)
    {
        $this->checkReport($entity);

        $clonedEntity = clone $entity;
        $clonedEntity->setName(
            $this->get(TranslatorInterface::class)->trans(
                'oro.report.action.clone.name_format',
                [
                    '{name}' => $clonedEntity->getName()
                ]
            )
        );

        return $this->update($clonedEntity);
    }

    /**
     * @Route(
     *      "/{_format}",
     *      name="oro_report_index",
     *      requirements={"_format"="html|json"},
     *      defaults={"_format" = "html"}
     * )
     *
     * @Template
     * @AclAncestor("oro_report_view")
     */
    public function indexAction()
    {
        return [];
    }

    /**
     * @param Report $entity
     * @return array
     */
    protected function update(Report $entity)
    {
        $reportForm = $this->get('form.factory')->createNamed(
            'oro_report_form',
            ReportFormType::class,
            $entity
        );
        $this->get(EntityNameProvider::class)->setCurrentItem($entity);
        if ($this->get(ReportHandler::class)->process($entity, $reportForm)) {
            $this->get('session')->getFlashBag()->add(
                'success',
                $this->get(TranslatorInterface::class)->trans('Report saved')
            );

            return $this->get(Router::class)->redirect($entity);
        }

        return [
            'entity'   => $entity,
            'form'     => $reportForm->createView(),
            'entities' => $this->get(EntityProvider::class)->getEntities(),
            'metadata' => $this->get(QueryDesignerManager::class)->getMetadata('report')
        ];
    }

    /**
     * Method detects type of report's chart 'label' field, and in case of datetime will check dates interval and
     * set proper type (time, day, date, month or year). Xaxis labels not taken into account - they will be rendered
     * automatically. Also chart dot labels may overlap if dates are close to each other.
     *
     * Should be refactored in scope of BAP-8294.
     *
     * @param DatagridInterface $datagrid
     * @param array             $chartOptions
     *
     * @return array
     */
    protected function processChartOptions(DatagridInterface $datagrid, array $chartOptions)
    {
        $labelFieldName = $chartOptions['data_schema']['label'];
        $labelFieldType = $datagrid->getConfig()->offsetGetByPath(
            sprintf('[columns][%s][frontend_type]', $labelFieldName)
        );

        /** @var DateHelper $dateTimeHelper */
        $dateTimeHelper = $this->get('oro_dashboard.datetime.helper');
        $dateTypes      = [Type::DATETIME, Type::DATE, Type::DATETIMETZ];

        if (in_array($labelFieldType, $dateTypes)) {
            $data  = $datagrid->getData()->offsetGet('data');
            $dates = array_map(
                function ($dataItem) use ($labelFieldName) {
                    return $dataItem[$labelFieldName];
                },
                $data
            );

            $minDate = new \DateTime(min($dates));
            $maxDate = new \DateTime(max($dates));

            $formatStrings = $dateTimeHelper->getFormatStrings($minDate, $maxDate);

            $chartOptions['data_schema']['label'] = [
                'field_name'   => $chartOptions['data_schema']['label'],
                'type'         => $formatStrings['viewType']
            ];
        }

        return $chartOptions;
    }

    /**
     * @param Report $report
     */
    protected function checkReport(Report $report)
    {
        if ($report->getEntity() &&
            !$this->get(FeatureChecker::class)->isResourceEnabled($report->getEntity(), 'entities')) {
            throw $this->createNotFoundException();
        }
    }
}
